import numpy as np
from copy import deepcopy
import mne
from scipy.sparse.csgraph import laplacian
from scipy.linalg import sqrtm, pinv
from scipy.sparse import csr_matrix
from scipy import sparse as sp
# from ..invert import BaseSolver, InverseOperator
# from .. import invert
from .base import BaseSolver, InverseOperator


class SolverLORETA(BaseSolver):
    ''' Class for the Low Resolution Tomography (LORETA) inverse solution.
    
    Attributes
    ----------

    References
    ----------
    [1] Pascual-Marqui, R. D. (1999). Review of methods for solving the EEG
    inverse problem. International journal of bioelectromagnetism, 1(1), 75-86.

    '''
    def __init__(self, name="Low Resolution Tomography", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha='auto', **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        leadfield = self.leadfield
        LTL = leadfield.T @ leadfield
        B = np.diag(np.linalg.norm(leadfield, axis=0))
        adjacency = mne.spatial_src_adjacency(forward['src'], verbose=self.verbose).toarray()
        laplace_operator = laplacian(adjacency)
        BLapTLapB = B @ laplace_operator.T @ laplace_operator @ B

   
        inverse_operators = []
        for alpha in self.alphas:
            inverse_operator = np.linalg.inv(LTL + (alpha) * BLapTLapB) @ leadfield.T
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self

class SolverSLORETA(BaseSolver):
    ''' Class for the standardized Low Resolution Tomography (sLORETA) inverse
        solution [1].
    
    Attributes
    ----------
    
    References
    ----------
    [1] Pascual-Marqui, R. D. (2002). Standardized low-resolution brain
    electromagnetic tomography (sLORETA): technical details. Methods Find Exp
    Clin Pharmacol, 24(Suppl D), 5-12.
    '''
    def __init__(self, name="Standardized Low Resolution Tomography", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha=0.01, verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        if alpha == "auto":
            msg = "sLORETA does not work well for automated regularization. Please use a floating points number (e.g., alpha=0.01)."
            raise AttributeError(msg)
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)

        leadfield = self.leadfield
        n_chans = leadfield.shape[0]
        
        LLT = leadfield @ leadfield.T
        
        I = np.identity(n_chans)
        # one = np.ones((n_chans, 1))
        # H = I - (one @ one.T) / (one.T @ one)

        inverse_operators = []
        for alpha in self.alphas:
            # according to Grech et al 2008
            K_MNE = leadfield.T @ np.linalg.pinv(LLT + alpha *I)
            W_diag = np.sqrt(np.diag(K_MNE @ leadfield))
            W_slor = (K_MNE.T / W_diag).T

            # according to pascual-marqui 2002
            # W_slor = leadfield.T @ H @ np.linalg.pinv(H @ LLT @ H + alpha * H)
            # J = leadfield.T @ np.linalg.pinv(LLT + alpha * H)
            # S = leadfield.T @ np.linalg.pinv(LLT + alpha * H) @ leadfield
            # W_slor = J.T @ np.linalg.inv(S) @ J
            # print(J.shape, S.shape, W_slor.shape)

            # According to pascual-marqui 2009 (?)
            # C = LLT + alpha*I
            # LTC = leadfield.T @ C
            # W_slor = np.linalg.pinv(LTC @ leadfield) @ LTC
            
            inverse_operator = W_slor
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self

class SolverELORETA(BaseSolver):
    ''' Class for the exact Low Resolution Tomography (eLORETA) inverse
        solution [1].
    
    Attributes
    ----------
    
    References
    ----------
    [1] Pascual-Marqui, R. D. (2007). Discrete, 3D distributed, linear imaging
    methods of electric neuronal activity. Part 1: exact, zero error
    localization. arXiv preprint arXiv:0710.3341.

    '''
    def __init__(self, name="Exact Low Resolution Tomography", **kwargs):
        self.name = name
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, *args, alpha='auto', verbose=0, stop_crit=0.005, max_iter=100, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        stop_crit : float
            The convergence criterion to optimize the weight matrix. 
        max_iter : int
            The stopping criterion to optimize the weight matrix.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        leadfield = self.leadfield
        n_chans = leadfield.shape[0]
        # noise_cov = np.identity(n_chans)
        
        # Some pre-calculations
        I = np.identity(n_chans)
        
        # No regularization leads to weird results with eLORETA
        if 0 in self.alphas and len(self.alphas) > 1:
            idx = self.alphas.index(0)
            self.alphas.pop(idx)
            self.r_values = np.delete(self.r_values, idx)
        elif 0 in self.alphas and len(self.alphas) == 1:
            idx = self.alphas.index(0)
            self.alphas = [0.01]
        
        inverse_operators = []
        for alpha in self.alphas:
            
            W = self.calc_W(alpha, max_iter=max_iter, stop_crit=stop_crit)
            W_inv = sp.linalg.inv(W)

            inverse_operator = W_inv @ leadfield.T @ pinv(leadfield @ W_inv @ leadfield.T + alpha * I)
            inverse_operators.append(inverse_operator)

        self.inverse_operators = [InverseOperator(inverse_operator, self.name) for inverse_operator in inverse_operators]
        return self

    def calc_W(self, alpha, max_iter=100, stop_crit=1e-3):
        from scipy.sparse import csr_matrix
        K = self.leadfield
        n_chans, n_dipoles= K.shape

        I = csr_matrix(np.identity(n_chans))
        W = csr_matrix(np.identity(n_dipoles))
        W_inv = W

        # Refine W iteratively
        for iter in range(max_iter):            
            W_old = deepcopy(W)
            W_inv = sp.linalg.inv(W)
                
            M = pinv(K @ W_inv @ K.T + alpha*I)

            W = csr_matrix(np.diag(np.sqrt(np.einsum('ij,jk,ki->i', K.T, M, K))))

            change = np.trace(abs(W.toarray()-W_old.toarray()))
            print(f"iter {iter}: {change}")
            if change < stop_crit:
                # converged!
                break

        return W

def calc_eloreta_D2(leadfield, noise_cov, alpha, stop_crit=0.005, verbose=0):
    ''' Algorithm that optimizes weight matrix D as described in 
        Assessing interactions in the brain with exactlow-resolution electromagnetic tomography; Pascual-Marqui et al. 2011 and
        https://www.sciencedirect.com/science/article/pii/S1053811920309150
        '''
    n_chans, n_dipoles = leadfield.shape
    # initialize weight matrix D with identity and some empirical shift (weights are usually quite smaller than 1)
    D = np.identity(n_dipoles)

    if verbose>0:
        print('Optimizing eLORETA weight matrix W...')
    cnt = 0
    while True:
        old_D = deepcopy(D)
        if verbose>0:
            print(f'\trep {cnt+1}')
        D_inv = np.linalg.inv(D)
        inner_term = np.linalg.inv(leadfield @ D_inv @ leadfield.T + alpha**2*noise_cov)
            
        for v in range(n_dipoles):
            D[v, v] = np.sqrt( leadfield[:, v].T @ inner_term @ leadfield[:, v] )
        
        averagePercentChange = np.abs(1 - np.mean(np.divide(np.diagonal(D), np.diagonal(old_D))))
        
        if verbose>0:
            print(f'averagePercentChange={100*averagePercentChange:.2f} %')

        if averagePercentChange < stop_crit:
            if verbose>0:
                print('\t...converged...')
            break
        cnt += 1
    if verbose>0:
        print('\t...done!')
    return D


from .base import BaseSolver, InverseOperator
import mne
import numpy as np
from tqdm import tqdm
from scipy.linalg import toeplitz
from scipy.sparse import spdiags, csr_matrix, diags

def pos_from_forward(forward, verbose=0):
    ''' Get vertex/dipole positions from mne.Forward model

    Parameters
    ----------
    forward : instance of mne.Forward
        The forward model. 
    
    Return
    ------
    pos : numpy.ndarray
        A 2D matrix containing the MNI coordinates of the vertices/ dipoles

    Note
    ----
    forward must contain some subject id in forward["src"][0]["subject_his_id"]
    in order to work.
    '''
    # Get Subjects ID
    subject_his_id = forward["src"][0]["subject_his_id"]
    src = forward["src"]

    # Extract vertex positions from left and right source space
    pos_left = mne.vertex_to_mni(src[0]["vertno"], 0, subject_his_id, 
                                verbose=verbose)
    pos_right = mne.vertex_to_mni(src[1]["vertno"], 1, subject_his_id, 
                                verbose=verbose)

    # concatenate coordinates from both hemispheres
    pos = np.concatenate([pos_left, pos_right], axis=0)

    return pos
class SolverMultipleSparsePriors(BaseSolver):
    ''' Class for the Multiple Sparse Priors (MSP) inverse solution.
    
    Attributes
    ----------
    forward : mne.Forward
        The mne-python Forward model instance.
    '''
    def __init__(self, name="Multiple Sparse Priors", inversion_type="MSP", **kwargs):
        if inversion_type == "MSP":
            self.name = name
        elif inversion_type == "LORETA":
            self.name = "Bayesian LORETA"
        elif inversion_type == "MNE":
            self.name = "Bayesian Minimum Norm Estimates"
        elif inversion_type == "BMF":
            self.name = "Bayesian Beamformer"
        elif inversion_type == "BMF-LOR":
            self.name = "Bayesian Beamformer + LORETA"
        self.inversion_type = inversion_type
        return super().__init__(**kwargs)

    def make_inverse_operator(self, forward, evoked, *args, Np=64, 
                              max_iter=128, smoothness=0.6, alpha='auto', 
                              verbose=0, **kwargs):
        ''' Calculate inverse operator.

        Parameters
        ----------
        forward : mne.Forward
            The mne-python Forward model instance.
        alpha : float
            The regularization parameter.
        
        Return
        ------
        self : object returns itself for convenience
        '''
        super().make_inverse_operator(forward, *args, alpha=alpha, **kwargs)
        leadfield = self.leadfield
        pos = pos_from_forward(forward, verbose=verbose)
        adjacency = mne.spatial_src_adjacency(forward['src'], verbose=verbose).toarray()
        Y = evoked.data
        A = get_spatial_projector(leadfield)
        
        S, V = get_temporal_projector(evoked, leadfield, A)
                
        Y_ = A @ Y @ S
        
        

        leadfield_ = A @ leadfield
        maximum_a_posteriori = make_msp_map(Y_, leadfield_, pos, adjacency, A, Np=Np, max_iter=max_iter, 
            inversion_type=self.inversion_type, smoothness=smoothness)
        inverse_operators = [maximum_a_posteriori, A, S]
        
        
        
        self.inverse_operators = [InverseOperator(inverse_operators, self.name),]
        return self

    def apply_inverse_operator(self, evoked) -> mne.SourceEstimate:
        S = self.inverse_operators[0].data[-1]
        if S.shape[1] != evoked.data.shape[1]:
            # print("\tRe-calculating projectors...")
            self.make_inverse_operator(self.forward, evoked)
            # print("\T...done!")
        return super().apply_inverse_operator(evoked)

def make_msp_inverse_operator(leadfield, pos, adjacency, evoked,Np=64, 
                              max_iter=128, inversion_type='MSP', 
                              smoothness=0.6, noise_cov=None, **kwargs):
    """ Calculate the inverse operator using Multiple Sparse Priors.

    Parameters
    ----------
    leadfield : numpy.ndarray
        Leadfield (or gain matrix) G which constitutes the forward model of M =
        J @ G, where sources J are projected through the leadfield producing the
        observed EEG matrix M.
    alpha : float
        The regularization parameter.
    noise_cov : numpy.ndarray
        The noise covariance matrix (channels x channels).

    Return
    ------
    inverse_operator : numpy.ndarray
        The inverse operator that is used to calculate source.

    """
    n_chans, _ = leadfield.shape
    if noise_cov is None:
        noise_cov = np.identity(n_chans)

    A = get_spatial_projector(leadfield)
    S, V = get_temporal_projector(evoked, leadfield, A)
    Y = evoked.data

    Y_ = A @ Y @ S
    leadfield_ = A @ leadfield
    maximum_a_posteriori = make_msp_map(Y_, leadfield_, pos, adjacency, A, Np=Np, max_iter=max_iter, 
        inversion_type=inversion_type, smoothness=smoothness)
    inverse_operator = [maximum_a_posteriori, A, S]
    # J_ = maximum_a_posteriori @ Y_
    # D_MSP =  J_ @ S.T 
    return inverse_operator


def inverse_msp(evoked, fwd, Np=64, max_iter=128, inversion_type='MSP', smoothness=0.6):
    leadfield, pos = unpack_fwd(fwd)[1:3]
    A = get_spatial_projector(leadfield)
    S, V = get_temporal_projector(evoked, leadfield, A)
    Y = evoked.data

    Y_ = A @ Y @ S
    leadfield_ = A @ leadfield
    maximum_a_posteriori = make_msp_map(Y_, leadfield_, fwd, A, Np=Np, max_iter=max_iter, 
        inversion_type=inversion_type, smoothness=smoothness)
    J_ = maximum_a_posteriori @ Y_
    D_MSP =  J_ @ S.T 
    return D_MSP



def get_spatial_projector(leadfield):
    # eliminate low SNR spatial modes
    thr = np.e**(-16)
    U,s2,v2    = np.linalg.svd(leadfield @ leadfield.T)
    # This replaces the custom svd function from spm called "spm_svd(...):"
    s2 *=  s2
    keep = (s2*len(s2) / s2.sum()) > thr
    # This replaces the custom svd function from spm called "spm_svd(...):"
    s2 = s2[keep]
    U = U[:, keep]
    v2 = v2[:,keep]
    del s2, v2

    A     = U.T;  # spatial projector A
    # UL    = np.matmul(A,L)
    # Nm    = UL.shape[0]				# Number of spatial projectors
    # Is = np.arange(Nd)
    # Ns = len(Is)
    return A

def get_temporal_projector(evoked, leadfield, A, Nmax=16, hpf=0, lpf=45, sdv=4):
    '''
    Parameters
    ----------
    Nmax : int
        max number of temporal modes
    '''
    Nd = leadfield.shape

    #  Time-window of interest
    It = np.arange(len(evoked.times)).astype(int)
    # print(It)
    # % Peristimulus time
    pst    = 1000*evoked.times  # peristimulus time (ms)
    pst    = pst[It]  # windowed time (ms)
    dur    = (pst[-1] - pst[0])/1000  # duration (s)
    dct    = (It - It[0])/2/dur  # DCT frequencies (Hz)
    Nb     = len(It)  # number of time bins
    # dct
    # # Serial correlations
    K      = np.e**(-(pst - pst[0])**2/(2*sdv**2));
    K      = toeplitz(K);
    qV     = np.matmul(K,K.T)
    # # Confounds and temporal subspace
    T      = spm_dctmtx(Nb,Nb)  # use plot(T) here!
    j      = (dct>=hpf) & (dct <=lpf)  # Filter
    T      = T[:,j]  # Apply the filter to discrete cosines
    dct    = dct[j]  # Frequencies accepted

    # No Hanning window
    # W  = sparse(1:Nb,1:Nb,spm_hanning(Nb));
    W  = 1  # Apply Hanning if desired!

    # get temporal covariance (Y'*Y) to find temporal modes
    # Note: The variable YY was replaced with YTY because it is
    # duplicated in the original script, causing confusion.
    D_tmp = evoked.data
    D_tmp = D_tmp[:, :]
    # print(D_tmp.shape)
    Y = np.matmul(A, D_tmp)  # Data samples in spatial modes (virtual sensors)
    YTY = np.matmul(Y.T,Y)  # Covariance in temporal domain

    # Apply any Hanning and filtering
    # YTY = np.matmul(np.matmul(W, YTY), W)  # Hanning
    YTY = np.matmul(np.matmul(T.T, YTY), T)  # Filter

    # Plot temporal projectors
    # plt.figure()
    # plt.imshow(YTY, extent=[dct.min(), dct.max(), dct.min(), dct.max()])  # Plot of frequency map
    # plt.title('Temporal projector')
    # plt.xlabel('Frequency (Hz)')
    # plt.ylabel('Frequency (Hz)')

    # temporal projector (at most Nrmax modes) S = T*V
    # [U E]  = spm_svd(YTY,exp(-8));		# get temporal modes
    U, E, _  = np.linalg.svd(YTY);		# get temporal modes
    thr = np.e**(-16)
    E *=  E
    keep = (E*len(E) / E.sum()) > thr
    # print(keep.sum()/len(keep), E)
    # keep = E>thr
    # U = U[:, keep]
    E = E[keep]

    # This replaces the custom svd function from spm called "spm_svd(...):"

    E      = np.diag(E)/np.trace(YTY)  # normalise variance
    Nr     = np.min([len(E),Nmax]) 	# number of temporal modes
    V      = U[:,:Nr]  # temporal modes
    VE     = np.sum(E[:Nr])  # variance explained

    # print(f'Using {Nr} temporal mode(s)',)
    # print(f'accounting for {100*VE:.1f} % average variance\n')

    # projection and whitening
    # print(YTY.shape, T.shape, Y.shape, V.shape)
    S      = T @ V  # temporal projector
    # Vq     = np.matmul(S, np.matmul(np.linalg.pinv(np.matmul(np.matmul(S.T, qV),S)), S.T))  # temporal precision

    # # stack (scaled aligned data) over modalities
    # Y    = np.matmul(np.matmul(A, D_tmp), S)  # Load data in spatial and temporal modes

    # # accumulate first & second-order responses
    # YY   = np.matmul(Y, Y.T)  # Data covariance in spatial mode

    # # generate sensor error components (Qe)
    # # assuming equal noise over subjects (Qe) and modalities AQ
    # AQeA   = np.matmul(A,A.T)  # Note that here it is A*A', it means
    # Qe  = AQeA/(np.trace(AQeA))  # IID noise in virtual sensor space
    return S, V

def greens_function(adjacency, smoothness=0.6):
    
    A = adjacency
    Nd = A.shape[0]
    GL = A - spdiags( A.sum(axis=1), 0, Nd, Nd)
    GL = GL * smoothness/2


    Qi = np.identity(Nd)
    QG = np.zeros((Nd, Nd))
    for i in range(8):
        QG += Qi
        Qi = Qi @ GL / (i+1)
    
    QG = QG * (QG > np.exp(-8))
    QG = QG @ QG
    return QG

def make_msp_map(Y, leadfield, pos, adjacency, A, Np=128, smoothness=0.6, max_iter=128,
    inversion_type='MSP'):
    '''
    Create the maximum a posteriori (MAP) estimator for the multiple sparse
    priors inverse solution.


    References
    ----------
    [1] Friston, K., Harrison, L., Daunizeau, J., Kiebel, S., Phillips, C.,
    Trujillo-Barreto, N., ... & Mattout, J. (2008). Multiple sparse priors for
    the M/EEG inverse problem. NeuroImage, 39(3), 1104-1120. 
    
    [2] López, J. D., Litvak, V., Espinosa, J. J., Friston, K., & Barnes, G. R.
    (2014). Algorithmic procedures for Bayesian MEG/EEG source reconstruction in
    SPM. NeuroImage, 84, 476-487.
    '''
    n, d = leadfield.shape
    # u, v = Y.shape
    C = Y @ Y.T  # sensor covariance
    # Scale C:
    sY = n*np.trace(C)
    C /= sY

    # v = 1
    # Greens function
    if smoothness is not None:
        Q_G = greens_function(adjacency, smoothness=smoothness)
    else:   
        Q_G = np.identity(d)
    
    # Get Sensor Covariance Components
    AQeA = A @ A.T
    sensor_noise = AQeA / np.trace(AQeA)
    Qe = [sensor_noise,]
    # Get Source Covariance Components
    Qp = []
    LQpL = []
    if inversion_type == 'MSP':
        Ip = np.ceil( np.arange(Np) * d/Np ).astype(int)
        m = deepcopy(len(Ip))  # numer of source cov components
        for i in range(m):
            # First Set
            Q = np.diag(Q_G[:, Ip[i]])
            Qp.append( Q )
            LQpL.append( leadfield@Q@leadfield.T )
            # Extended Set (other hemi)
            j = np.argmin(np.sum(
                np.array([
                    pos[:, 0] + pos[Ip[i], 0],
                    pos[:, 1] - pos[Ip[i], 1], 
                    pos[:, 2] - pos[Ip[i], 2]])**2, axis=0))
            Q = np.diag(Q_G[:, j])
            Qp.append(Q)
            LQpL.append( leadfield@Q@leadfield.T )
            # bilateral
            Q = np.diag(Q_G[:, Ip[i]]) + np.diag(Q_G[:, j])
            Qp.append(Q)
            LQpL.append( leadfield@Q@leadfield.T )
            
    elif inversion_type == 'MNE':
        Qp.append( np.identity(d) )
        LQpL.append( leadfield @ leadfield.T )
            
    elif inversion_type == 'LORETA':
        Qp.append( np.identity(d) )
        LQpL.append( leadfield @ leadfield.T )

        Qp.append( Q_G )
        LQpL.append( leadfield @ Q_G @ leadfield.T )
        
    elif inversion_type == 'BMF':
        inv_cov = np.linalg.inv(C)
        Ns = pos.shape[0]
        allsource = np.zeros(Ns)
        sourcepower = np.zeros(Ns)
        for bk in range(Ns):
            ll = leadfield[:, bk][:, np.newaxis]
            normpower = 1 / (ll.T  @ ll)
            sourcepower[bk] = 1 / (ll.T @  inv_cov @ ll)
            allsource[bk] = sourcepower[bk]  / normpower
        allsource /= allsource.max()
        Qp.append( np.diag(allsource) )
        LQpL.append( leadfield @ np.diag(allsource) @ leadfield.T )

    elif inversion_type == 'BMF-LOR':
        inv_cov = np.linalg.inv(C)
        Ns = pos.shape[0]
        allsource = np.zeros(Ns)
        sourcepower = np.zeros(Ns)
        for bk in range(Ns):
            ll = leadfield[:, bk][:, np.newaxis]
            normpower = 1 / (ll.T  @ ll)
            sourcepower[bk] = 1 / (ll.T @  inv_cov @ ll)
            allsource[bk] = sourcepower[bk]  / normpower
        allsource /= allsource.max()

        Qp.append( np.diag(allsource) )
        LQpL.append( leadfield @ np.diag(allsource) @ leadfield.T )
        
        Qp.append( Q_G )
        LQpL.append( leadfield @ Q_G @ leadfield.T )
        

    m = len(Qp)
    assert len(LQpL) == len(Qp), "Source Covs and projected source covs are not equal."
    

    
    # Source Level Analysis
    # M-STEP
        
    # Combine the noise covariance with the projected source covariances
    Q = np.array(Qe + LQpL)
    # for q in Q:
    #     print(q.shape)
    if inversion_type == "MSP":
        h, _ = spm_sp_reml_demo(C, Q, max_iter)
    else:
        Q0          = np.exp(-2) * np.trace(C) * Qe[0] / np.trace(Qe[0])
        _, h, _, _, _, _ = spm_reml_sc_demo(C, Qe + LQpL, 1, -4, 16, Q0);



    # Ne = len(Qe)
    # Np = len(Qp)
    # # hp = h[Ne + np.arange(Np)]
    # hp = h[Ne:]
    # qp = 0
    # for i in range(Np):
    #     if hp[i] > np.max(hp) / 128:
    #         qp  += hp[i] * Qp[i]
    
    # QP = [np.diag(qp),]
    # LQP = [leadfield @ qp,]
    # LQPL = [LQP[-1] @ leadfield.T]

    # AQ    = AQeA / np.trace(AQeA)
    # Np = len(LQPL)
    # Ne = len(Qe)
    # Q = np.array(Qe + LQPL)

    # Q0 = np.exp(-2) * np.trace(C) * AQ / np.trace(AQ)
    # Cy, h, _, _, _, _ = spm_reml_sc_demo(C, Q, 1, -4, 16, Q0)
    # Cp    = 0
    # LCp   = 0
    # Ne = 1
    # hp    = h[Ne + np.arange(Np)]
    # for j in range(Np):
    #     # print(hp[j], QP[j].shape, LQP[j])
    #     Cp  +=  hp[j] * QP[j]
    #     LCp +=  hp[j] * LQP[j]
    
    
    # M     = LCp.T / Cy 
    
    # E-Step
    sigma_e = sigma(h[1:], Qp)
    sigma_u_lam = sigma(h, Q)
    M = sigma_e @ leadfield.T @ np.linalg.inv(sigma_u_lam)
    
    return M


def spm_sp_reml_demo(YY, Q_c, max_iter):
    n = YY.shape[0]
    m = len(Q_c)
    h = np.zeros(m)  # hyperparameters
    
    
    hE = np.zeros(m) - 32
    hC = np.identity(m) * 256
    hP = np.linalg.inv(hC)

    # Scale
    sY = n*np.trace(YY)
    YY  /= sY

    # find bases of Q if necessary
    Q = []
    for i in range(m):
        q, v, _ = np.linalg.svd(Q_c[i])
        thr = np.exp(-16)
        v *=  v
        keep = (v*len(v) / v.sum()) > thr
        q = q[:, keep]
        v = v[keep]
        C = dict(q=q, v=v)
        Q.append(C)



    # Scaling
    sh = np.zeros(m);
    for i in range(m):
        sh[i] = n * np.trace(Q[i]["q"].T @ Q[i]["q"] @ np.diag(Q[i]["v"]));
        Q[i]["q"]  = Q[i]["q"] / np.sqrt(sh[i]);
    
    q = [qq["q"] for qq in Q]
    v = [qq["v"] for qq in Q]

    a = 0
    for i in range(len(v)):
        a += len(v[i])
    
    dedh = np.zeros((a, len(v)))
    # print("dedh.shape: ", dedh.shape, "v[0].shape: ", v[0].shape, "v[1].shape: ", v[1].shape, )
    dedh[:n, 0] = v[0]
    for i in range(1, dedh.shape[1]):
        b = np.zeros(a)
        b[i+n-1:i+n-1+len(v[i])] = v[i]
        dedh[:, i] = b

    

    delta_F = 1
    a_s = np.arange(m)
    for ii in range(max_iter):
        # C = np.zeros((n, n))
        # e = dedh * np.exp(h)
        # s = len(q)
        # print(len(q), q[0].shape, e.shape)
        # for i in range(s):
        #     C += q[i] @ e[i]

        sigma_u_lam = sigma(h, Q_c)
        sigma_u_pinv = np.linalg.pinv(sigma_u_lam)
        e = np.exp(h)

        W = sigma_u_pinv @ YY @ sigma_u_pinv - sigma_u_pinv
        qr = np.stack([np.diagonal(Q) for Q in Q_c[a_s]], axis=1)
        P = qr.T @ sigma_u_pinv @ qr
        
        dFde = 0.5 * np.sum(qr * (W@qr), axis=0).T
        # print(dFde.shape)
        dFdee = -0.5 * P * P.T

        dhdh = np.diag(np.exp(h[a_s]))
        dFdh = dhdh.T @ dFde
        
        dFdhh = dhdh.T @ dFdee @ dhdh

        e = h - hE
        dFdh = dFdh - hP[a_s, a_s] * e[a_s]
        
        dFdhh = dFdhh - hP[a_s, a_s]

        dh = -np.linalg.pinv(dFdhh) @ dFdh / np.log(ii+3)
        h[a_s] += dh

        dF = dFdh.T @ dh

        # print(f'Iteration {ii+1}. Free Energy Improvement: {dF:.2f}')        
        if (delta_F < 0.01) or (ii == max_iter):
            break
        else:
            a_s = h > -16
            h[~a_s] = hE[~a_s]
            h[h>1] = 1
        
      
    h  = sY*np.exp(h) / sh
    return h, a_s

def sigma(X, Q_e):
        return np.sum([np.e**x * Q for x, Q in zip(X, Q_e)], axis=0)

def spm_dctmtx(N,K):
    # Create basis functions for Discrete Cosine Transform
    # translated and simplified from:
    #  https://github.com/spm/spm12/blob/master/spm_dctmtx.m
    
    n = np.arange(N).T
        

    C = np.zeros((len(n), K))
    # print(np.ones((n.shape[0],1)).shape)
    C[:, 0] = np.ones(n.shape[0]) / np.sqrt(N)

    for k in range(1,K):
        C[:,k] = np.sqrt(2/N)*np.cos(np.pi*(2*n+1)*(k-1)/(2*N))
    return C

def spm_mvb_G_demo(Y,QL,G,Qe):
    # Demo version of the spm_mvb_G_demo() routine
    # Related with the Technical Note:
    # 
    # Algorithmic procedures for Bayesian MEG/EEG source reconstruction in SPM
    # 
    # Created by:	Jose David Lopez - ralph82co@gmail.com
    # 				Gareth Barnes - g.barnes@fil.ion.ucl.ac.uk
    # 				Vladimir Litvak - litvak.vladimir@gmail.com
    # 
    # Date: April 2013


    # 1. INITIALISE DEFAULTS

    # defaults
    Nk    = G.shape[0]
    Np    = G.shape[1]
    Ne    = len(Qe)
    
    # assemble empirical priors
    Qp    = []
    LQpL  = []
    for i in range(Np):
        j = np.where(G[:,i])[0];
        # mat = sparse(j,j,1,Nk,Nk)
        mat = np.zeros((Nk, Nk))
        mat[j, j] = 1
        Qp.append( mat )
        new_mat = np.matmul(np.matmul(QL,mat), QL.T)
        LQpL.append( new_mat ) 
    
    
    # 2. INVERSE SOLUTION
    # Covariance components (with the first error Qe{1} fixed)
    
    if QL.shape[1] > 0:
        Q = [Qe, LQpL]
    else:
        Q = Qe
    
    # ReML (with mildly informative hyperpriors) and a lower bound on error
    N          = Y.shape[1]
    YY         = np.matmul(Y,Y.T)
    Qe         = (np.e**(-2))*np.mamtul(np.trace(YY),Q[0]) / (np.trace(Q[0])/N)
    hE         = -4
    hC         = 16
    # TODO implement
    [Cy,h,Ph,F] = spm_reml_sc_demo(YY,Q,N,hE,hC,Qe)

    # 3. ESTIMATE AND SAVE

    # prior covariance: source space
    Cp    = np.zeros((Nk,Nk))
    for i in range(Np):
        Cp += np.matmul(h[i + Ne], Qp[i])
    
    # MAP estimates of instantaneous sources
    MAP   = np.matmul(Cp,QL.T/Cy)
    qE    = np.matmul(MAP,Y)
    
    
    # assemble results (pattern weights)
    d = dict(F=F, G=G, h=h, qE=qE, MAP=MAP, Cp=Cp)
    return d

def spm_reml_sc_demo(YY,Q,N,hE,hC,Qe):
    # Demo version of the spm_reml_sc_demo() routine
    # Related with the Technical Note:
    #
    # Algorithmic procedures for Bayesian MEG/EEG source reconstruction in SPM
    # 
    # Created by:	Jose David Lopez - ralph82co@gmail.com
    #				Gareth Barnes - g.barnes@fil.ion.ucl.ac.uk
    #				Vladimir Litvak - litvak.vladimir@gmail.com
    #
    # Date: April 2013
    #

    # 1. INITIALISE DEFAULTS

    # initialise h
    n     = Q[0].shape[0]
    m     = len(Q)
    h     = np.zeros(m)
    dFdh  = np.zeros(m)
    dFdhh = np.zeros((m,m))
    
    # check fixed component
    if len(Qe) == 1:
        Qe = np.matmul(Qe,np.eye(n,n))
    
    # initialise and specify hyperpriors
    # scale Q and YY
    sY = np.trace(YY)/(N*n)
    YY = YY / sY
    Qe  = Qe / sY
    sh = np.zeros(m)
    for i in range(m):
        sh[i] = np.trace(Q[i]) / n
        Q[i]    = Q[i]/sh[i]
    

    # hyperpriors
    try:
        hP = np.linalg.inv(hC)
    except:
        hP = 1/hC
    
    
    hE = hE*np.ones(m)

    hP = hP*np.identity(m)

    # intialise h: so that sum(exp(h)) = 1
    if np.any(np.diag(hP) > np.e**(16)):
        h = hE
    
    # 2. ReML (EM/VB)
    dF    = np.inf
    a_s    = np.arange(m)
    t     = 4
    for k in range(32):

        # 2.1 UPDATE PRIOR COVARIANCE MATRIX
        # E-step: conditional covariance cov(B|y) {Cq}
        # compute current estimate of covariance
        C     = Qe
        for i in a_s:
            C += Q[i]*np.exp(h[i])
        P    = np.linalg.inv(C)
        # M-step: ReML estimate of hyperparameters
        # 2.2 COMPUTE GRADIENT AND CURVATURE
        # Gradient dF/dh (first derivatives)
        U     = np.eye(n) - P @ YY / N
        PQ = dict()
        for i in a_s:
            # dF/dh = -trace(dF/diC*iC*Q{i}*iC)
            PQ[i] = P @ Q[i]
            dFdh[i] = -np.trace( PQ[i] @ U) * N/2
        

        # Expected curvature E{dF/dhh} (second derivatives)
        for i in a_s:
            for j in a_s:
                # dF/dhh = -trace{P*Q{i}*P*Q{j}}
                dFdhh[i,j] = -np.trace( PQ[i] @ PQ[j]) * N/2
                dFdhh[j,i] =  dFdhh[i,j]
  
        # 2.3 UPDATE HYPERPARAMETERS
        # modulate
        dFdh  = dFdh * np.exp(h)
        
        dFdhh = np.dot(dFdhh,  np.exp(h)[:, np.newaxis] @ np.exp(h)[np.newaxis])
    
        # add hyperpriors
        e     = h     - hE
        dFdh  = dFdh  - (hP @ e[:, np.newaxis])[:, 0]
        
        dFdhh = dFdhh - hP

        # Fisher scoring: update dh = -inv(ddF/dhh)*dF/dh
        # ToDo: bring "t" into the equation
        dh = -np.linalg.pinv(dFdhh) @ dFdh
        # dh    = spm_dx(dFdhh[a_s,a_s], dFdh[a_s], [t])
        h[a_s] += dh[a_s]

        # predicted change in F - increase regularisation if increasing
        pF    = dFdh[a_s].T @ dh[a_s]
        if pF > dF:
            t -= 1
        else:
            t += 1/8;
        dF    = pF
        
        # print(f"ReML Iteration {k}: {dF}") 
        # print(f'%s %-23d: %10s%e [%+3.2f]\n','  ReML Iteration',k,'...',full(dF),t);
        if dF < 1e-2:
            break
        else:
            # eliminate redundant components (automatic selection)
            a_s  = np.where(h > -16)[0]
            h[~a_s] = hE[~a_s]
            h[a_s] = 1
            # a_s  = a_s.T
        
    # 3. COMPUTATION OF FREE ENERGY

    # log evidence = ln p(y|X,Q) = ReML objective = F = trace(R'*iC*R*YY)/2 ...
    Ph    = -dFdhh
    
    # tr(hP*inv(Ph)) - nh + tr(pP*inv(Pp)) - np (pP = 0)
    # Ft = np.trace(hP/Ph) - len(Ph)
    Ft = 0
    # complexity - KL(Ph,hP)
    # Fc = Ft/2 + np.e*hP*np.e/2 + np.log(np.linalg.det( Ph/hP )) / 2
    Fc = 0
    # Accuracy - ln p(Y|h)
    # Fa = Ft/2 - np.trace(C*P*YY*P)/2 - N*n*np.log(2*np.pi)/2  - N * np.log(np.linalg.det(C))/2
    Fa = 0
    # Free-energy
    # F  = Fa - Fc - N*n*np.log(sY)/2;
    F = 0
    # print('Free-energy: ', F)
    
    

    # return exp(h) hyperpriors and rescale
    # h  = np.log(sY*np.exp(h) / sh)
    # print("final h: ", h)
    C  = sY*C;
    return C,h,Ph,F,Fa,Fc
    


import numpy as np
from .util import pos_from_forward
import matplotlib.pyplot as plt

from invert import solvers
from . import config


def Solver(solver="mne", **kwargs):
    """ Solver class ...
    """
    # Minimum Norm Algorithms
    if solver.lower() == "mne":
        solver_object = solvers.SolverMNE(**kwargs)
    elif solver.lower() == "wmne":
        solver_object = solvers.SolverWMNE(**kwargs)
    elif solver.lower() == "dspm":
        solver_object = solvers.SolverDSPM(**kwargs)
    elif solver.lower() == "l1" or solver.lower() == "fista" or solver.lower() == "mce" or solver.lower() == "minimum current estimate":
        solver_object = solvers.SolverMinimumL1Norm(**kwargs)
    elif solver.lower() == "gpt" or solver.lower() == "l1 gpt" or solver.lower() == "l1gpt":
        solver_object = solvers.SolverMinimumL1NormGPT(**kwargs)
    elif solver.lower() == "l1l2":
        solver_object = solvers.SolverMinimumL1L2Norm(**kwargs)
    
    
    # LORETA Algorithms
    elif solver.lower() == "loreta" or solver.lower() == "lor":
        solver_object = solvers.SolverLORETA(**kwargs)
    elif solver.lower() == "sloreta" or solver.lower() == "slor":
        solver_object = solvers.SolverSLORETA(**kwargs)
    elif solver.lower() == "eloreta" or solver.lower() == "elor":
        solver_object = solvers.SolverELORETA(**kwargs)
    
    # Various smooth Algorithms
    elif solver.lower() == "laura" or solver.lower() == "laur":
        solver_object = solvers.SolverLAURA(**kwargs)
    elif solver.lower() == "backus-gilbert" or solver.lower() == "b-g" or  solver.lower() == "bg":
        solver_object = solvers.SolverBackusGilbert(**kwargs)
    elif solver.lower() == "s-map" or solver.lower() == "smap":
        solver_object = solvers.SolverSMAP(**kwargs)
    
    # Bayesian Algorithms
    elif solver.lower() == "champagne" or solver.lower() == "champ":
        solver_object = solvers.SolverChampagne(**kwargs)
    elif solver.lower() == "lowsnrchampagne" or solver.lower() == "low snr champagne" or solver.lower() == "lowsnr champagne"  or solver.lower() == "lsc":
        solver_object = solvers.SolverChampagne(**kwargs, update_rule="LowSNR")
    elif solver.lower() == "mackaychampagne" or solver.lower() == "mackay champagne" or solver.lower() == "mackay-champagne" or solver.lower() == "mcc":
        solver_object = solvers.SolverChampagne(**kwargs, update_rule="MacKay")
    elif solver.lower() == "convexitychampagne" or solver.lower() == "convexity champagne" or solver.lower() == "convexity-champagne" or solver.lower() == "coc":
        solver_object = solvers.SolverChampagne(**kwargs, update_rule="Convexity")
    elif solver.lower() == "champagne-nl" or solver.lower() == "champagne nl" or solver.lower() == "nl champagne":
        solver_object = solvers.SolverNLChampagne(**kwargs)

    elif solver.lower() == "emchampagne" or solver.lower() == "em champagne" or solver.lower() == "emc":
        solver_object = solvers.SolverEMChampagne(**kwargs)
    elif solver.lower() == "mmchampagne" or solver.lower() == "mm champagne" or solver.lower() == "mmc":
        solver_object = solvers.SolverMMChampagne(**kwargs)
    elif solver.lower() == "fun":
        solver_object = solvers.SolverFUN(**kwargs)
    elif solver.lower() == "heteroscedastic champagne" or solver.lower() == "hs champagne" or solver.lower() == "hs-champagne" or solver.lower() == "hsc":
        solver_object = solvers.SolverHSChampagne(**kwargs)
    # elif solver.lower() == "temchampagne" or solver.lower() == "tem champagne" or solver.lower() == "temc":
    #     solver_object = solvers.SolverTEMChampagne(**kwargs)
    elif solver.lower() == "gamma-map" or solver.lower() == "gamma map" or solver.lower() == "gmap":
        solver_object = solvers.SolverGammaMAP(**kwargs)
    elif solver.lower() == "source-map" or solver.lower() == "source map":
        solver_object = solvers.SolverSourceMAP(**kwargs)
    elif solver.lower() == "gamma-map-msp" or solver.lower() == "gamma map msp":
        solver_object = solvers.SolverGammaMAPMSP(**kwargs)
    elif solver.lower() == "source-map-msp" or solver.lower() == "source map msp":
        solver_object = solvers.SolverSourceMAPMSP(**kwargs)
    
    # Beamformer Algorithms
    elif solver.lower() == "mvab":
        solver_object = solvers.SolverMVAB(**kwargs)
    elif solver.lower() == "lcmv":
        solver_object = solvers.SolverLCMV(**kwargs)
    elif solver.lower() == "smv":
        solver_object = solvers.SolverSMV(**kwargs)
    elif solver.lower() == "wnmv":
        solver_object = solvers.SolverWNMV(**kwargs)
    elif solver.lower() == "hocmv":
        solver_object = solvers.SolverHOCMV(**kwargs)
    elif solver.lower() == "esmv":
        solver_object = solvers.SolverESMV(**kwargs)
    elif solver.lower() == "mcmv":
        solver_object = solvers.SolverMCMV(**kwargs)
    elif solver.lower() == "hocmcmv":
        solver_object = solvers.SolverHOCMCMV(**kwargs)
    # elif solver.lower() == "esmcmv":
    #     solver_object = solvers.SolverESMCMV(**kwargs)
    elif solver.lower() == "recipsiicos":
        solver_object = solvers.SolverReciPSIICOS(**kwargs)
    elif solver.lower() == "sam":
        solver_object = solvers.SolverSAM(**kwargs)
    elif solver.lower() == "ebb" or solver.lower() == "empirical bayesian beamformer":
        solver_object = solvers.SolverEBB(**kwargs)

    # Own approaches
    elif solver.lower() == "fully-connected" or solver.lower() == "fc" or solver.lower() == "fullyconnected" or solver.lower() == "esinet":
        solver_object = solvers.SolverFC(**kwargs)
    elif solver.lower() == "covcnn" or solver.lower() == "cov cnn" or solver.lower() == "covnet" or solver.lower() == "cov-cnn":
        solver_object = solvers.SolverCovCNN(**kwargs)
    elif solver.lower() == "lstm":
        solver_object = solvers.SolverLSTM(**kwargs)
    elif solver.lower() == "cnn":
        solver_object = solvers.SolverCNN(**kwargs)
    
    # Matching Pursuit/ Compressive Sensing
    elif solver.lower() == "omp":
        solver_object = solvers.SolverOMP(**kwargs)
    elif solver.lower() == "cosamp":
        solver_object = solvers.SolverCOSAMP(**kwargs)
    elif solver.lower() == "somp":
        solver_object = solvers.SolverSOMP(**kwargs)
    elif solver.lower() == "rembo":
        solver_object = solvers.SolverREMBO(**kwargs)
    elif solver.lower() == "sp":
        solver_object = solvers.SolverSP(**kwargs)
    elif solver.lower() == "ssp":
        solver_object = solvers.SolverSSP(**kwargs)
    elif solver.lower() == "smp":
        solver_object = solvers.SolverSMP(**kwargs)
    elif solver.lower() == "ssmp":
        solver_object = solvers.SolverSSMP(**kwargs)
    elif solver.lower() == "subsmp":
        solver_object = solvers.SolverSubSMP(**kwargs)
    elif solver.lower() == "isubsmp":
        solver_object = solvers.SolverISubSMP(**kwargs)
    elif solver.lower() == "bcs":
        solver_object = solvers.SolverBCS(**kwargs)
    
    # Subspace Methods
    elif solver.lower() == "music":
        solver_object = solvers.SolverMUSIC(**kwargs)
    elif solver.lower() == "rap-music" or solver.lower() == "rap music" or solver.lower() == "rap":
        solver_object = solvers.SolverFLEXMUSIC(name="RAP-MUSIC", **kwargs)
    elif solver.lower() == "trap-music" or solver.lower() == "trap music" or solver.lower() == "trap":
        solver_object = solvers.SolverFLEXMUSIC(name="TRAP-MUSIC", **kwargs)
    elif solver.lower() == "flex-rap-music" or solver.lower() == "flex-music" or solver.lower() == "flex music" or solver.lower() == "flex":
        solver_object = solvers.SolverFLEXMUSIC(**kwargs)
    elif solver.lower() == "flex-ssm" or solver.lower() == "flex ssm":
        solver_object = solvers.SolverSignalSubspaceMatching(name="Flexible Signal Subspace Matching", **kwargs)
    elif solver.lower() == "ssm" or solver.lower() == "signal subspaces matching":
        solver_object = solvers.SolverSignalSubspaceMatching(name="Signal Subspace Matching", **kwargs)
    
    # Alternative Projections
    elif solver.lower() == "flex-ap" or solver.lower() == "flex ap":
        solver_object = solvers.SolverAlternatingProjections(name="Flexible Alternating Projections", **kwargs)
    elif solver.lower() == "ap":
        solver_object = solvers.SolverAlternatingProjections(name="Alternating Projections", **kwargs)

    # Basis Functions
    elif solver.lower() == "gbf":
        solver_object = solvers.SolverBasisFunctions(**kwargs)
        
    # Other
    elif solver.lower() == "epifocus":
        solver_object = solvers.SolverEPIFOCUS(**kwargs)

    
    
    
    else:
        msg = f"{solver} is not available. Please choose from one of the following: {config.all_solvers}"
        raise AttributeError(msg)
    return solver_object





from copy import deepcopy
import numpy as np
import colorednoise as cn
from scipy.sparse.csgraph import laplacian
from scipy.sparse import csr_matrix, vstack
import mne
import pandas as pd


def generator(fwd, use_cov=True, cov_type="basic", batch_size=1284, batch_repetitions=30, n_sources=10, 
              n_orders=2, amplitude_range=(0.001,1), n_timepoints=20, 
              snr_range=(-5, 5), n_timecourses=5000, beta_range=(0, 3),
              return_mask=True, scale_data=True, return_info=False,
              add_forward_error=False, forward_error=0.1, remove_channel_dim=False, 
              inter_source_correlation=0.5, diffusion_smoothing=True, 
              diffusion_parameter=0.1, correlation_mode=None, 
              noise_color_coeff=0.5, random_seed=None, 
              normalize_leadfield=False, verbose=0):
    """
    Parameters
    ----------
    fwd : object
        Forward solution object containing the source space and orientation information.
    use_cov : bool
        If True, a covariance matrix is used in the simulation. Default is True.
    batch_size : int
        Size of each batch of simulations. Default is 1284.
    batch_repetitions : int
        Number of repetitions of each batch. Default is 30.
    n_sources : int
        Number of sources in the brain from which activity is simulated. Default is 10.
    n_orders : int
        The order of the model used to generate time courses. Default is 2.
    amplitude_range : tuple
        Range of possible amplitudes for the simulated sources. Default is (0.001,1).
    n_timepoints : int
        Number of timepoints in each simulated time course. Default is 20.
    snr_range : tuple
        Range of signal to noise ratios (in dB) to be used in the simulations. Default is (-5, 5 dB).
    n_timecourses : int
        Number of unique time courses to simulate. Default is 5000.
    beta_range : tuple
        Range of possible power-law exponents for the power spectral density of the simulated sources. Default is (0, 3).
    return_mask : bool
        If True, the function will also return a mask of the sources used. Default is True.
    scale_data : bool
        If True, the EEG data will be scaled. Default is True.
    return_info : bool
        If True, the function will return a dictionary with information about the generated data. Default is False.
    add_forward_error : bool
        If True, the function will add an error to the forward model. Default is False.
    forward_error : float
        Amount of error to add to the forward model if 'add_forward_error' is True. Default is 0.1.
    remove_channel_dim : bool
        If True, the channel dimension will be removed from the output data. Default is False.
    inter_source_correlation : float|Tuple
        The level of correlation between different sources. Default is 0.5.
    diffusion_smoothing : bool
        Whether to use diffusion smoothing. Default is True.
    diffusion_parameter : float
        The diffusion parameter (alpha). Default is 0.1.
    correlation_mode : None/str
        correlation_mode : None/str
        None implies no correlation between the noise in different channels.
        'banded' : Colored banded noise, where channels closer to each other will be more correlated.
        'diagonal' : Channels have varying degrees of noise.
    noise_color_coeff : float
        The magnitude of spatial coloring of the noise.
    random_seed : None / int
        The random seed for replicable simulations
    verbose : int
        Level of verbosity for the function. Default is 0.
    
    Return
    ------
    x : numpy.ndarray
        The EEG data matrix.
    y : numpy.ndarray
        The source data matrix.
    """
    rng = np.random.default_rng(random_seed)
    channel_types = fwd["info"].get_channel_types()
    leadfield = deepcopy(fwd["sol"]["data"])
    leadfield_original = deepcopy(fwd["sol"]["data"])
    n_chans, n_dipoles = leadfield.shape

    if isinstance(n_sources, (int, float)):
        n_sources = [np.clip(n_sources, a_min=1, a_max=np.inf), n_sources]
    min_sources, max_sources = n_sources

    adjacency = mne.spatial_src_adjacency(fwd["src"], verbose=verbose)
    # Convert to sparse matrix for speedup
    adjacency = csr_matrix(adjacency)
    if diffusion_smoothing:
        # print("Using Diffusion Smoothing on Graph")
        gradient = np.identity(n_dipoles) - diffusion_parameter*laplacian(adjacency)
    else:
        gradient = abs(laplacian(adjacency))
    
    # Convert to sparse matrix for speedup
    gradient = csr_matrix(gradient)

    del adjacency
    
    
    # leadfield -= leadfield.mean()
    # Normalize columns of the leadfield
    if normalize_leadfield:
        leadfield /= np.linalg.norm(leadfield, axis=0)

    sources = csr_matrix(np.identity(n_dipoles))
    if isinstance(n_orders, (tuple, list)):
        min_order, max_order = n_orders
        if min_order == max_order:
            max_order += 1
    else:
        min_order = 0
        max_order = n_orders

    if isinstance(inter_source_correlation, (tuple, list)):
        get_inter_source_correlation = lambda n=1: rng.uniform(inter_source_correlation[0], inter_source_correlation[1], n)
    else:
        get_inter_source_correlation = lambda n=1: rng.uniform(inter_source_correlation, inter_source_correlation, n)

    if isinstance(noise_color_coeff, (tuple, list)):
        get_noise_color_coeff = lambda n=1: rng.uniform(noise_color_coeff[0], noise_color_coeff[1], n)
    else:
        get_noise_color_coeff = lambda n=1: rng.uniform(noise_color_coeff, noise_color_coeff, n)

    if type(correlation_mode) == str and correlation_mode.lower() == "auto":
        correlation_modes = ["cholesky", "banded", "diagonal", None]
        correlation_modes = rng.choice(correlation_modes, batch_size)
    else:
        correlation_modes = [correlation_mode,] * batch_size

    # print(get_inter_source_correlation())

    for i in range(1, max_order):
        # new_sources = sources[-n_dipoles:, -n_dipoles:] @ gradient
        # new_sources /= new_sources.max(axis=0)
        # sources = np.concatenate( [sources, new_sources], axis=0 )
        
        # New Smoothness order
        new_sources = csr_matrix(sources.toarray()[-n_dipoles:, -n_dipoles:]) @ gradient
        
        # Scaling
        row_maxes = new_sources.max(axis=0).toarray().flatten()
        new_sources = new_sources / row_maxes[np.newaxis]

        sources = vstack([sources, new_sources])
    # print(sources.shape)
    if min_order > 0:
        start = int(min_order*n_dipoles)
        sources = sources.toarray()[start:, :]
    
    sources = csr_matrix(sources)
    # Pre-compute random time courses
    betas = rng.uniform(*beta_range,n_timecourses)
    random_states = rng.integers(0, 2**32, n_timecourses)
    time_courses = np.stack([cn.powerlaw_psd_gaussian(beta, n_timepoints, random_state=random_state) for beta, random_state in zip(betas, random_states)], axis=0)
    
    # Normalize time course to max(abs()) == 1
    time_courses = (time_courses.T / abs(time_courses).max(axis=1)).T

    n_candidates = sources.shape[0]
    # print(sources.shape, n_orders)
    while True:
        if add_forward_error:
            leadfield = add_error(leadfield_original, forward_error, gradient, rng)
        # print("yeet")
        # select sources or source patches
        n_sources_batch = rng.integers(min_sources, max_sources+1, batch_size)
        selection = [rng.integers(0, n_candidates, n) for n in n_sources_batch]

        # Assign each source (or source patch) a time course
        # amplitude_values = [rng.uniform(*amplitude_range, n) for n in n_sources_batch]
        # amplitudes = [time_courses[rng.choice(n_timecourses, n)].T * amplitude_values[i] for i, n in enumerate(n_sources_batch)]

        amplitude_values = [rng.uniform(*amplitude_range, n) for n in n_sources_batch]
        choices = [rng.choice(n_timecourses, n) for n in n_sources_batch]
        # print(choices)
        amplitudes = [time_courses[choice].T for choice in choices]
        from scipy.stats import pearsonr
        # print(len(amplitudes), amplitudes[0].shape)
        # print("Initial corr between two timecourses: ", pearsonr(amplitudes[0][0], amplitudes[0][1])[0])
        inter_source_correlations = get_inter_source_correlation(n=batch_size)
        if type(noise_color_coeff) != str:
            noise_color_coeffs = get_noise_color_coeff(n=batch_size)
        else:
            noise_color_coeffs = [noise_color_coeff,] * batch_size
        source_covariances = [get_cov(n, isc) for n, isc in zip(n_sources_batch, inter_source_correlations)]
        # print(source_covariances)
        amplitudes = [amp @ np.diag(amplitude_values[i]) @ cov for i, (amp, cov) in enumerate(zip(amplitudes, source_covariances))]

        # print("All the data: \n")
        # print("n_sources_batch: ", n_sources_batch, np.min(n_sources_batch), np.max(n_sources_batch))
        # print("selection: ", selection)
        # print("amplitude_values: ", amplitude_values)
        # print("inter_source_correlations: ", inter_source_correlations)
        

        # print(np.stack(amplitudes, axis=0).shape)

        # y = np.stack([(amplitudes[i] @ sources.toarray()[selection[i]]) / len(amplitudes[i]) for i in range(batch_size)], axis=0)
        y = np.stack([(amplitudes[i] @ sources[selection[i]]) / len(amplitudes[i]) for i in range(batch_size)], axis=0)

        # Project simulated sources through leadfield
        x = np.stack([leadfield @ yy.T for yy in y], axis=0)
        # Add white noise to clean EEG
        snr_levels = rng.uniform(low=snr_range[0], high=snr_range[1], size=batch_size)
        x = np.stack([add_white_noise(xx, snr_level, rng, channel_types, correlation_mode=correlation_mode, noise_color_coeff=noise_color_level) for (xx, snr_level, correlation_mode, noise_color_level) in zip(x, snr_levels, correlation_modes, noise_color_coeffs)], axis=0)

        # Apply common average reference
        # x = np.stack([xx - xx.mean(axis=0) for xx in x], axis=0)
        
        if use_cov:
            # Calculate Covariance
            
            x = np.stack([compute_covariance(xx, cov_type=cov_type) for xx in x], axis=0)
            
            # Normalize Covariance to abs. max. of 1
            x = np.stack([C / np.max(abs(C)) for C in x], axis=0)
            # x = np.stack([C / (np.trace(C) / C.shape[0]) for C in x], axis=0)

            if not remove_channel_dim:
                x = np.expand_dims(x, axis=-1)
        
        else:
            if scale_data:
                # normalize all time points to unit length
                # x = np.stack([xx / np.linalg.norm(xx, axis=0) for xx in x], axis=0)
                # normalize each sample to max(abs()) == 1
                x = np.stack([xx / np.max(abs(xx)) for xx in x], axis=0)
                # x = np.stack([xx / (np.trace(xx@xx.T) / n_chans) for xx in x], axis=0)
            # Reshape
            x = np.swapaxes(x, 1,2)

        if return_mask:
            # (1) binary
            # Calculate mean source activity
            y = abs(y).mean(axis=1)
            # Masking the source vector (1-> active, 0-> inactive)
            y = (y>0).astype(float)
            
            # (2.1) Source Covariance
            # The diagonal of the source covariance matrix:
            # y = np.stack([np.diagonal(yy.T@yy) for yy in y], axis=0)

            # (2.2) Source Covariance (efficient)
            # y = np.mean(y**2, axis=1)
            # y = np.stack([yy / np.max(abs(yy)) for yy in y], axis=0)
        
        else:
            if scale_data:
                y = np.stack([ (yy.T / np.max(abs(yy), axis=1)).T for yy in y], axis=0)
        
        if return_info:
            info = pd.DataFrame(dict(
                n_sources=n_sources_batch, 
                amplitudes=amplitude_values, 
                snr=snr_levels, 
                inter_source_correlations=inter_source_correlations, 
                n_orders=[[min_order, max_order],]*batch_size,
                diffusion_parameter=[diffusion_parameter,] * batch_size,
                n_timepoints=[n_timepoints,] * batch_size,
                n_timecourses=[n_timecourses,] * batch_size,
                correlation_mode=[correlation_mode,] * batch_size,
                noise_color_coeff=noise_color_coeffs,
                centers=selection
                ))
            output = (x, y, info)
        else:
            output = (x, y)

        for _ in range(batch_repetitions):
            yield output


def compute_covariance(Y, cov_type="basic"):
        ''' Compute the covariance matrix of the data.

        Parameters
        ----------
        Y : numpy.ndarray
            The data matrix.
        cov_type : str
            The type of covariance matrix to compute. Options are 'basic' and 'SSM'. Default is 'basic'.
        
        Return
        ------
        C : numpy.ndarray
            The covariance matrix.
        '''
        if cov_type == "basic":
            C = Y @ Y.T
        elif cov_type == "SSM":
            n_time = Y.shape[1]
            M_Y = Y.T @ Y
            YY = M_Y + 0.001 * (50/n_time) * np.trace(M_Y) * np.eye(n_time)
            P_Y = (Y @ np.linalg.inv(YY)) @ Y.T
            C = P_Y.T @ P_Y
        else:
            msg = "Covariance type not recognized. Use 'basic', 'SSM' or provide a custom covariance matrix."
            raise ValueError(msg)
        
        return C

def generator_simple(fwd, batch_size, corrs, T, n_sources, SNR_range, 
                     random_seed=42, return_info=True):
        rng = np.random.default_rng(random_seed)
        leadfield = deepcopy(fwd["sol"]["data"])
        leadfield /= leadfield.std(axis=0)
        n_channels, n_dipoles = leadfield.shape

        while True:
            sim_info = list()
            X = np.zeros((batch_size, n_channels, T))
            y = np.zeros((batch_size, n_dipoles, T))
            corrs_batch = rng.uniform(corrs[0], corrs[1], batch_size)
            SNR_batch = rng.uniform(SNR_range[0], SNR_range[1], batch_size)
            indices = [rng.choice(fwd["sol"]["data"].shape[1], n_sources) for _ in range(batch_size)]
            
            for i in range(batch_size):
                # print(leadfield.shape, corrs_batch[i], T, n_sources, indices[i], SNR_batch[i], random_seed)
                X[i], y[i] = generator_single_simple(leadfield, corrs_batch[i], T, n_sources, indices[i], SNR_batch[i], random_seed=random_seed)
                d = dict(
                    n_sources=n_sources, 
                    amplitudes=1,
                    snr=SNR_batch[i],
                    inter_source_correlations=corrs_batch[i],
                    n_orders = [0,0],
                    diffusion_parameter=0,
                    n_timepoints=T,
                    n_timecourses=np.inf,
                    iid_noise=True)
                sim_info.append(d)
            if return_info:
                sim_info = pd.DataFrame(sim_info)
                yield X, y, sim_info
            else:
                yield X, y


def generator_single_simple(leadfield, corr, T, n_sources, indices, SNR, random_seed=42):
    """
    Parameters
    ----------
    leadfield : numpy.ndarray
        The leadfield matrix.
    corr : float
        The correlation coefficient between the sources.
    T : int
        The number of time points in the sources.
    n_sources : int
        The number of sources to generate.
    indices : list
        The indices of the sources to generate.
    SNR : float
        The signal to noise ratio.
    random_seed : int
        The random seed for replicable simulations.
    
    Return
    ------
    X : numpy.ndarray
        The simulated EEG data. 
    y: numpy.ndarray
        The simulated source data.
    """
    rng = np.random.default_rng(random_seed)
    
    S = gen_correlated_sources(corr, T, n_sources)
    M = leadfield[:, indices] @ S # use Ground Truth Gain matrix
    n_channels, n_dipoles = leadfield.shape

    scale = np.max(abs(M))
    Ms = M * scale
    MEG_energy = np.trace(Ms @ Ms.T) / (n_channels*T)
    noise_var = MEG_energy/(10**(SNR/10))
    Noise = rng.standard_normal((n_channels, T)) * np.sqrt(noise_var)
    X = Ms + Noise
    y = np.zeros((n_dipoles, T))
    y[indices, :] = S
    
    
    return X, y

def gen_correlated_sources(corr_coeff, T, Q):
    ''' Generate Q correlated sources with a specified correlation coefficient.
    The sources are generated as sinusoids with random frequencies and phases.

    Parameters
    ----------
    corr_coeff : float
        The correlation coefficient between the sources.
    T : int
        The number of time points in the sources.
    Q : int
        The number of sources to generate.

    Returns
    -------
    Y : numpy.ndarray
        The generated sources.
    '''
    Cov = np.ones((Q, Q)) * corr_coeff + np.diag(np.ones(Q) * (1 - corr_coeff))  # required covariance matrix
    freq = np.random.randint(10, 31, Q)  # random frequencies between 10Hz to 30Hz

    phases = 2 * np.pi * np.random.rand(Q)  # random phases
    t = np.linspace(10 * np.pi / T, 10 * np.pi, T)
    Signals = np.sqrt(2) * np.cos(2 * np.pi * freq[:, None] * t + phases[:, None])  # the basic signals

    if corr_coeff < 1:
        A = np.linalg.cholesky(Cov).T  # cholesky Decomposition
        Y = A @ Signals
    else:  # Coherent Sources
        Y = np.tile(Signals[0, :], (Q, 1))

    return Y


def get_cov(n, corr_coef):
    '''Generate a covariance matrix that is symmetric along the
    diagonal that correlates sources to a specified degree.'''
    if corr_coef < 1:
        cov = np.ones((n,n)) * corr_coef + np.eye(n)*(1-corr_coef)
        cov = np.linalg.cholesky(cov)
    else:
        # Make all signals be exactly the first one (perfectly coherent)
        cov = np.zeros((n, n))
        cov[:, 0] = 1
    return cov.T

def add_white_noise(X_clean, snr, rng, channel_types, noise_color_coeff=0.5, correlation_mode=None):
    ''' 
    Parameters
    ----------
    X_clean : numpy.ndarray
        The clean EEG data.
    snr : float
        The signal to noise ratio in dB.
    correlation_mode : None/str
        None implies no correlation between the noise in different channels.
        'banded' : Colored banded noise, where channels closer to each other will be more correlated.
        'diagonal' : Some channels have varying degrees of noise.
        'cholesky' : A set correlation coefficient between each pair of channels
    noise_color_coeff : float
        The magnitude of spatial coloring of the noise (not the magnitude of noise overall!).
    '''
    n_chans, n_time = X_clean.shape
    X_noise = rng.standard_normal((n_chans, n_time))
    snr_linear = 10**(snr/10)
    
    if type(channel_types == list):
        channel_types = np.array(channel_types)
    # Ensure the channel_types array is correct length
    assert len(channel_types) == n_chans, "Length of channel_types must match the number of channels in X_clean"
    
    unique_types = np.unique(channel_types)
    X_full = np.zeros_like(X_clean)
    
    for ch_type in unique_types:
        
        type_indices = np.where(channel_types == ch_type)[0]
        # print(len(type_indices), ch_type)
        X_clean_type = X_clean[type_indices, :]
        X_noise_type = X_noise[type_indices, :]
        if (type(noise_color_coeff) == str) & (type(correlation_mode) == np.ndarray):
            # Real Noise Covariance
            # X_noise_type = correlation_mode[type_indices,type_indices] @ X_noise_type
            X_noise_type = np.linalg.cholesky(correlation_mode[type_indices][:,type_indices]) @ X_noise_type
        elif correlation_mode == "cholesky":
            covariance_matrix = np.full((len(type_indices), len(type_indices)), noise_color_coeff)
            np.fill_diagonal(covariance_matrix, 1)  # Set diagonal to 1 for variance
            
            # mean = np.zeros(len(type_indices))  # Mean of the noise
            # X_noise_type = np.random.multivariate_normal(mean, covariance_matrix, n_time).T

            # Cholesky decomposition
            X_noise_type = np.linalg.cholesky(covariance_matrix) @ X_noise_type

            
            
            # import matplotlib.pyplot as plt
            # plt.figure()
            # plt.plot(X_noise_type.T)
            
            # C_noise = X_noise_type @ X_noise_type.T
            # C_noise_r = C_noise / np.sqrt(np.diag(C_noise)[:, None] @ np.diag(C_noise)[None, :])
            
            # import matplotlib.pyplot as plt
            
            # plt.figure()
            # plt.subplot(1,2,1)
            # plt.imshow(C_noise, vmin=-np.max(abs(C_noise)), vmax=np.max(abs(C_noise)), cmap="coolwarm")
            # plt.title(f"Covariance Matrix of {ch_type} noise")
            # plt.colorbar()  

            # plt.subplot(1,2,2)
            # plt.imshow(C_noise_r, vmin=-1, vmax=1, cmap="coolwarm")
            # plt.title(f"Correlation Matrix of {ch_type} noise")
            # plt.colorbar()

            
        elif correlation_mode == "banded":
            num_sensors = X_noise_type.shape[0]
            Y = np.zeros_like(X_noise_type)
            for i in range(num_sensors):
                Y[i, :] = X_noise_type[i, :]
                for j in range(num_sensors):
                    if abs(i - j) % num_sensors == 1:
                        Y[i, :] += (noise_color_coeff / np.sqrt(2)) * X_noise_type[j, :]
            X_noise_type = Y
            # C_noise = X_noise_type @ X_noise_type.T
            # C_noise /= np.sqrt(np.diag(C_noise)[:, None] @ np.diag(C_noise)[None, :])
        elif correlation_mode == "diagonal":
            X_noise_type[1::3, :] *= (1 - noise_color_coeff)
            X_noise_type[2::3, :] *= (1 + noise_color_coeff)
            # C_noise = X_noise_type @ X_noise_type.T
            # C_noise /= np.sqrt(np.diag(C_noise)[:, None] @ np.diag(C_noise)[None, :])
        elif correlation_mode is None:
            pass
        else:
            msg = f"correlation_mode can be either None, cholesky, banded or diagonal, but was {correlation_mode}"
            raise AttributeError(msg)
        
        rms_noise = rms(X_noise_type)
        rms_signal = rms(X_clean_type)
        scaler = rms_signal / (snr_linear * rms_noise)
        # print that the noise is scaled
        # effective_snr = rms(X_clean_type) / rms(X_noise_type * scaler)
        # print(f"Effective SNR for {ch_type} is {effective_snr}")


        X_full[type_indices] = X_clean_type + X_noise_type * scaler
        

        # # Amirs version
        # X_clean_energy = np.trace(X_clean_type @ X_clean_type.T) / (X_clean_type.shape[0] * X_clean_type.shape[1])
        # noise_var = X_clean_energy / snr_linear
        # scaler = np.sqrt(noise_var)
        # X_full[type_indices, :] = X_clean_type + X_noise_type * scaler
        # effective_snr = rms(X_clean_type) / rms(X_noise_type * scaler)
        # print(f"Effective SNR for {ch_type} is {effective_snr}")

    # print(f"effectice snr overall is {rms(X_clean) / rms(X_full - X_clean)}")
    return X_full

# def add_white_noise(X_clean, snr, rng, noise_color_coeff=0.5, correlation_mode=None):
#     ''' 
#     Parameters
#     ----------
#     correlation_mode : None/str
#         None implies no correlation between the noise in different channels.
#         'banded' : Colored banded noise, where channels closer to each other will be more correlated.
#         'diagonal' : Some channels have varying degrees of noise.
#         'cholesky' : A set correlation coefficient between each pair of channels
#     noise_color_coeff : float
#         The magnitude of spatial coloring of the noise (not the magnitude of noise overall!).
#     '''
#     # print(X_clean.shape)
#     n_chans, n_time = X_clean.shape
#     X_noise = rng.standard_normal((n_chans, n_time))
#     # X_noise = rng.uniform(0, 1, (n_chans, n_time))
#     # Ensure equal noise variance
#     # X_noise = (X_noise.T / np.var(X_noise, axis=1)).T
    
#     if correlation_mode == "cholesky":

#         covariance_matrix = np.full((n_chans, n_chans), noise_color_coeff)
#         np.fill_diagonal(covariance_matrix, 1)  # Set diagonal to 1 for variance
        
#         # Generate correlated noise
#         mean = np.zeros(n_chans)  # Mean of the noise
#         X_noise = np.random.multivariate_normal(mean, covariance_matrix, n_time).T
    
#         C_noise = X_noise @ X_noise.T
#         C_noise /= np.sqrt(np.diag(C_noise)[:, None] @ np.diag(C_noise)[None, :])
#     elif correlation_mode == "banded":
#         num_sensors = X_noise.shape[0]
#         Y = np.zeros_like(X_noise)
        
#         # Apply coloring to the noise
#         for i in range(num_sensors):
#             Y[i, :] = X_noise[i, :]
#             for j in range(num_sensors):
#                 if abs(i - j) % num_sensors == 1:
#                     Y[i, :] += (noise_color_coeff / np.sqrt(2)) * X_noise[j, :]
#         X_noise = Y

#         C_noise = X_noise @ X_noise.T
#         C_noise /= np.sqrt(np.diag(C_noise)[:, None] @ np.diag(C_noise)[None, :])

#     elif correlation_mode == "diagonal":
#         # Apply coloring to the noise
#         X_noise[1::3, :] *= (1 - noise_color_coeff)
#         X_noise[2::3, :] *= (1 + noise_color_coeff)

#         C_noise = X_noise @ X_noise.T
#         C_noise /= np.sqrt(np.diag(C_noise)[:, None] @ np.diag(C_noise)[None, :])
#     elif correlation_mode is None:
#         pass
#     else:
#         msg = f"correlation_mode can bei either None, cholesky, banded or diagonal, but was {correlation_mode}"
#         raise AttributeError(msg)
    
#     # rms_noise = rms(X_noise)
#     # rms_signal = rms(X_clean)

#     # scaler = rms_signal / (snr * rms_noise)
    
#     # # According to Adler et al.:
#     X_clean_energy = np.trace(X_clean@X_clean.T)/(X_clean.shape[0]*X_clean.shape[1])
#     noise_var = X_clean_energy/snr
#     scaler = np.sqrt(noise_var)

#     X_full = X_clean + X_noise*scaler
#     return X_full

def add_error(leadfield, forward_error, gradient, rng):
    n_chans, n_dipoles = leadfield.shape
    noise = rng.uniform(-1, 1, (n_chans, n_dipoles)) @ gradient
    leadfield_mix = leadfield / np.linalg.norm(leadfield) + forward_error * noise / np.linalg.norm(noise)
    return leadfield_mix

def rms(x):
    return np.sqrt(np.mean(x**2))
    

